package org.bdware.ypkdeploy;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import org.apache.http.HttpEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.bdware.client.ResultCallback;
import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.pqc.math.linearalgebra.ByteUtils;
import org.zz.gmhelper.SM2KeyPair;
import org.zz.gmhelper.SM2Util;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.atomic.AtomicInteger;

public class HTTPTool {
    final static String TAG = "[HTTPTool] ";

    static class DeployConfig {
        //common config
        String host;
        String ypkPath;
        String privateKey;
        String publicKey;
        String killBeforeStart;
        JsonElement createParam;
        //config for http deploy
        int agentPort;
    }


    public static void batchRun(String path, boolean restart) {
        DeployConfig config = null;
        try {
            config = new Gson().fromJson(new FileReader(path), DeployConfig.class);
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println(TAG + " parse config failed!");
        }
        deployUseHttp(config);
        SM2KeyPair keyPair = SM2KeyPair.fromJson(new Gson().toJson(config));
        SmartContractClientExt ext = new SmartContractClientExt(String.format("ws://%s:%d/SCIDE/SCExecutor", config.host, config.agentPort), keyPair);
        ext.waitForConnect();
        ext.login();
        for (; !ext.isLoggedIn; )
            Thread.yield();
        AtomicInteger counter = new AtomicInteger(0);

        try {

            if (config.killBeforeStart != null)
                ext.kill(config.killBeforeStart, new ResultCallback() {
                    @Override
                    public void onResult(JsonObject r) {
                        System.out.println(TAG + r);
                        counter.incrementAndGet();
                    }
                });
            for (; counter.get() == 0; )
                Thread.yield();
            ;
            Thread.sleep(200);
        } catch (Exception e) {
            e.printStackTrace();
        }
        File f = new File(config.ypkPath);
        ext.startContract(f.getName()
                .replaceAll(".zip", "")
                .replaceAll(".ypk", ""), config.createParam, new ResultCallback() {
            @Override
            public void onResult(JsonObject r) {
                System.out.println(TAG + r);
                counter.incrementAndGet();
            }
        });
        for (; counter.get() == 1; )
            Thread.yield();
        ;
    }


    public static void deployUseHttp(DeployConfig config) {
        String url = "http://%s:%d/Upload?%s&sign=%s";
        File file = new File(config.ypkPath);
        String path = config.ypkPath;
        String argWithoutSig = "path=%s&fileName=%s&isPrivate=true&order=%d&count=%d&pubKey=%s";
        SM2KeyPair keyPair = SM2KeyPair.fromJson(String.format("{\"publicKey\":\"%s\",\"privateKey\":\"%s\"}", config.publicKey, config.privateKey));
        String pubKey = keyPair.getPublicKeyStr();
        int trunc = 490 * 1024;
        byte[] buff = new byte[trunc];
        int order = 0;
        int count = (int) ((file.length() - 1) / (trunc)) + 1;
        try {
            FileInputStream fin = new FileInputStream(file);
            for (int len = 0; (len = fin.read(buff)) > 0; order++) {
                CloseableHttpClient client = HttpClients.createDefault();
                String arg = String.format(argWithoutSig, "./", file.getName(), order, count, pubKey);
                String sign = ByteUtils.toHexString(SM2Util.sign(keyPair.getPrivateKeyParameter(), arg.getBytes(StandardCharsets.UTF_8)));
                String urlStr = String.format(url, config.host, config.agentPort, arg, sign);
                HttpPost httpPost = new HttpPost(urlStr);
                MultipartEntityBuilder builder = MultipartEntityBuilder.create();
                if (len == trunc)
                    builder.addBinaryBody(
                            "file", buff,
                            ContentType.APPLICATION_OCTET_STREAM, file.getName());
                else {
                    byte[] bu = new byte[len];
                    System.arraycopy(buff, 0, bu, 0, len);
                    builder.addBinaryBody(
                            "file", bu,
                            ContentType.APPLICATION_OCTET_STREAM, file.getName());

                }
                HttpEntity multipart = builder.build();
                httpPost.setEntity(multipart);
                CloseableHttpResponse response = client.execute(httpPost);
                ByteArrayOutputStream bo = new ByteArrayOutputStream();
                response.getEntity().writeTo(bo);
                System.out.println("[YpkDeployTool] " + order
                        + " " + count + " " + bo);
            }
        } catch (IOException | CryptoException e) {
            e.printStackTrace();
        }
    }

    public static void main(String[] args) {
        batchRun("./deployconfig.json", true);
    }
}
